from __future__ import print_function

import sys
import time

from django.core.management.base import BaseCommand

from optparse import make_option

from chroniker.models import Job, Log, JobDependency

from criticalpath import Node

class Command(BaseCommand):
    help = 'Calculates the total time a series of chained jobs will take.'
    
    option_list = BaseCommand.option_list + (
#        make_option('--seconds',
#            dest='seconds',
#            default=60,
#            help='The number of total seconds to count up to.'),
        make_option('--samples',
            default=20,
            help='The number of log samples to use when estimating mean job run time.'),
        )
    
    def handle(self, root_job_id, **options):
        root_job = Job.objects.get(id=int(root_job_id))
        samples = int(options['samples'])
        
        # Add all system task nodes.
        system = Node('system')
        system.add(Node(root_job.id, duration=root_job.get_run_length_estimate(samples=samples)))
        print('%s takes about %s seconds' % (root_job, root_job.get_run_length_estimate(samples=samples)))
        chain = root_job.get_chained_jobs()
        for job in chain:
            print('%s takes about %s seconds' % (job, job.get_run_length_estimate(samples=samples)))
            node = Node(job.id, duration=job.get_run_length_estimate(samples=samples))
            node.description = job.name
            system.add(node)
        
        # Add all links between task nodes.
        print('-'*80)
        for job in chain:
            if not job.enabled:
                continue
            dependees = JobDependency.objects.filter(dependent=job, dependee__enabled=True).values_list('dependee_id', flat=True)
            print(job, dependees)
            for dependee in dependees:
                # Link dependent job to dependee.
                assert job.id != 1
                system.link(from_node=dependee, to_node=job.id)
        
        root_node = system.lookup_node(1)
        print('root_node:',root_node,root_node.to_nodes,root_node.incoming_nodes)
        system.add_exit()
        sys.stdout.flush()
        
        #return
        print('Updating values...')
        system.update_all()
        
        critical_path = system.get_critical_path()
        print('critical_path:',critical_path)
        system.print_times()
        
        print('min hours:',system.duration*(1/60.)*(1/60.))
        


from __future__ import print_function
import sys
import time

from django.core.management.base import BaseCommand, CommandError

class Sleeper(BaseCommand):
    args = '[time in seconds to loop]'
    help = 'A simple command that simply sleeps for the specified duration'
    
    def handle(self, target_time, **options):
        start_time = time.time()
        target_time = float(target_time)
        
        print("Sleeping for {} seconds...".format(target_time))
        time.sleep(target_time)
        
        end_time = time.time()
        print("Job ran for {} seconds".format(end_time-start_time))

class InfiniteWaiter(BaseCommand):
    args = ''
    help = 'A simple command that waits indefinitely.'
    
    def handle(self, *args, **options):
        while 1:
            time.sleep(1)
            print('Waiting...')
            
class ErrorThrower(BaseCommand):
    args = ''
    help = 'A simple command that always fails.'
    
    def handle(self, *args, **options):
        print(
            'Something went wrong (but not really, this is just a test).',
            file=sys.stderr)




from django import template
from django.core.urlresolvers import reverse, NoReverseMatch
from django.utils import timezone
from django.conf import settings

register = template.Library()

class RunJobURLNode(template.Node):
    def __init__(self, object_id):
        self.object_id = template.Variable(object_id)
        
    def render(self, context):
        object_id = self.object_id.resolve(context)
        try:
            # Old way
            url = reverse('chroniker_job_run', args=(object_id,))
        except NoReverseMatch:
            # New way
            url = reverse('admin:chroniker_job_run', args=(object_id,))
        return url

def do_get_run_job_url(parser, token):
    """
    Returns the URL to the view that does the 'run_job' command.
    
    Usage::
    
        {% get_run_job_url [object_id] %}
    """
    try:
        # Splitting by None == splitting by spaces.
        tag_name, object_id = token.contents.split(None, 1)
    except ValueError:
        raise template.TemplateSyntaxError(
            "%r tag requires one argument" % token.contents.split()[0])
    return RunJobURLNode(object_id)

register.tag('get_run_job_url', do_get_run_job_url)

@register.simple_tag
def now_offset(format_string, offset_days=0):
    """
    Like Django's built-in {% now ... %} tag, except it accepts a second
    integer parameter representing days to add to the current datetime.
    """
    from django.template.defaultfilters import date
    from datetime import datetime, timedelta
    offset_days = int(offset_days)
    tzinfo = timezone.get_current_timezone() if settings.USE_TZ else None
    dt = datetime.now(tz=tzinfo) + timedelta(days=offset_days)
    return date(dt, format_string)
    


from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import time
import signal
import subprocess
import warnings
import errno

from datetime import timedelta
#from StringIO import StringIO
try:
    from io import StringIO
except ImportError:
    from cStringIO import StringIO

from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import reverse
from django.db import models
from django.db import connection
from django.utils import timezone
from django.conf import settings

from multiprocessing import Process, current_process

import psutil

import six
from six import print_

from . import constants as c

def get_etc(complete_parts, total_parts, start_datetime, current_datetime=None, as_seconds=False):
    """
    Estimates a job's expected time to completion.
    """
    current_datetime = current_datetime or timezone.now()
    
    # Calculate the seconds passed since we started tracking progress.
    passed_seconds = float((current_datetime - start_datetime).total_seconds())
    
    # Estimate the total seconds the task will take to complete by using
    # a linear projection.
    total_seconds = passed_seconds/complete_parts*total_parts
    
    # Estimate the expected time of completion by projecting the duration
    # onto the start time.
    etc = start_datetime + timedelta(seconds=total_seconds)
    
    # If we only want remaining seconds, return difference between ETC and
    # the current time in seconds.
    if as_seconds:
        return (etc - current_datetime).total_seconds()
        
    return etc

def get_remaining_seconds(*args, **kwargs):
    kwargs['as_seconds'] = True
    return get_etc(*args, **kwargs)

#def calculate_eta(start_datetime, start_count, current_count, total_count):
#    """
#    Returns the datetime when the given process will likely complete, assuming
#    a relatively linear projection of the current progress.
#    """
#    assert start_count >= 0, 'Invalid start_count: %s' % (start_count,)
#    assert current_count >= 0, 'Invalid current_count: %s' % (current_count,)
#    assert total_count >= 0, 'Invalid total_count: %s' % (total_count,)
#    assert isinstance(start_datetime, datetime)
#    if not total_count:
#        return
#    now_datetime = datetime.now()
#    ran_parts = current_count - start_count
#    ran_seconds = (now_datetime - start_datetime).total_seconds()
#    remaining_parts = total_count - current_count - start_count
#    if not ran_parts:
#        return
#    remaining_seconds = ran_seconds/float(ran_parts)*remaining_parts
#    eta = now_datetime + timedelta(seconds=remaining_seconds)
#    return eta

def get_admin_change_url(obj):
    ct = ContentType.objects.get_for_model(obj)
    change_url_name = 'admin:%s_%s_change' % (ct.app_label, ct.model)
    return reverse(change_url_name, args=(obj.id,))

def get_admin_changelist_url(obj):
    ct = ContentType.objects.get_for_model(obj)
    list_url_name = 'admin:%s_%s_changelist' % (ct.app_label, ct.model)
    return reverse(list_url_name)

class TeeFile(StringIO):
    """
    A helper class for allowing output to be stored in a StringIO instance
    while still be directed to a second file object, such as sys.stdout.
    """
    def __init__(self, file, auto_flush=False, queue=None):
        #super(TeeFile, self).__init__()
        StringIO.__init__(self)
        self.file = file
        self.auto_flush = auto_flush
        self.length = 0
        self.queue = queue
        self.queue_buffer = []

    def write(self, s):
        try:
            #import chardet
            #encoding_opinion = chardet.detect(s)
            #encoding = encoding_opinion['encoding']
            #TODO:fix? not stripping out non-ascii characters result in error
            #'ascii' codec can't encode character ? in position ?: ordinal not in range(128)
            s = ''.join(_ for _ in s if ord(_) < 128)
            #s = s.encode(encoding, 'ignore')
        except ImportError:
            pass
        self.length += len(s)
        self.file.write(s)
        #super(TeeFile, self).write(s)
        StringIO.write(self, s)
        if self.auto_flush:
            #self.file.flush()
            self.flush()
        if self.queue is not None:
            self.queue_buffer.append(s)

    def flush(self):
        self.file.flush()
        #super(TeeFile, self).flush()
        StringIO.flush(self)
        if self.queue is not None:
            self.queue.put((current_process().pid, ''.join(self.queue_buffer)))
            self.queue_buffer = []

# Based on:
# http://djangosnippets.org/snippets/833/
# http://www.shiningpanda.com/blog/2012/08/08/mysql-table-lock-django/
class LockingManager(models.Manager):
    """ Add lock/unlock functionality to manager.

    Example::

        class Job(models.Model):

            manager = LockingManager()

            counter = models.IntegerField(null=True, default=0)

            @staticmethod
            def do_atomic_update(job_id)
                ''' Updates job integer, keeping it below 5 '''
                try:
                    # Ensure only one HTTP request can do this update at once.
                    Job.objects.lock()

                    job = Job.object.get(id=job_id)
                    # If we don't lock the tables two simultanous
                    # requests might both increase the counter
                    # going over 5
                    if job.counter < 5:
                        job.counter += 1
                        job.save()

                finally:
                    Job.objects.unlock()


    """

    def lock(self):
        """ Lock table.

        Locks the object model table so that atomic update is possible.
        Simulatenous database access request pend until the lock is unlock()'ed.

        Note: If you need to lock multiple tables, you need to do lock them
        all in one SQL clause and this function is not enough. To avoid
        dead lock, all tables must be locked in the same order.

        See http://dev.mysql.com/doc/refman/5.0/en/lock-tables.html
        """
        cursor = connection.cursor()
        if 'mysql' in connection.settings_dict['ENGINE']:
            table = self.model._meta.db_table
            logger.debug("Locking table %s" % table)
            cursor.execute("LOCK TABLES %s WRITE" % table)
        else:
            warnings.warn(
                'Locking of database backend "%s" is not supported.'
                    % (connection.settings_dict['ENGINE'],),
                warnings.RuntimeWarning
            )
        #row = cursor.fetchone()
        #return row
        return cursor

    def unlock(self):
        """ Unlock the table. """
        cursor = connection.cursor()
        if 'mysql' in connection.settings_dict['ENGINE']:
            table = self.model._meta.db_table
            cursor.execute("UNLOCK TABLES")
        else:
            warnings.warn(
                '(Un)Locking of database backend "%s" is not supported.'
                    % (connection.settings_dict['ENGINE'],),
                warnings.RuntimeWarning
            )
        #row = cursor.fetchone()
        #return row
        return cursor

def pid_exists(pid):
    """
    Returns true if the process associated with the given PID is still running.
    Returns false otherwise.
    """
    pid = int(pid)
    if pid < 0:
        return False
    try:
        os.kill(pid, 0)
    except OSError as e:
        return e.errno == errno.EPERM
    else:
        return True

def get_cpu_usage(pid, interval=1):
    """
    Returns the CPU usage, as reported by `ps`, of the process associated with
    the given PID.
    """
#    cmd = ['ps', '-p', str(pid), '-o', '%cpu', '--no-headers']
#    output = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()[0]
#    try:
#        return float(output.strip().split('\n')[0])
#    except ValueError:
#        return
    return psutil.Process(pid).get_cpu_times(interval=interval)

def kill_process(pid):
    """
    Kills the process associated with the given PID.
    Returns true if the process was successfully killed.
    Returns false otherwise.
    """
    pid = int(pid)
    try:

        # Try sending a keyboard interrupt.
        os.kill(pid, signal.SIGINT) # 2
        if not pid_exists(pid):
            return True

        # Ask politely again.
        os.kill(pid, signal.SIGABRT) # 6
        if not pid_exists(pid):
            return True

        # Try once more.
        os.kill(pid, signal.SIGTERM) # 15
        if not pid_exists(pid):
            return True

        # We've asked nicely and we've been ignored, so just murder it.
        os.kill(pid, signal.SIGKILL) # 9
        if not pid_exists(pid):
            return True

    except OSError:
        # Something strange happened.
        # Our user likely doesn't have permission to kill the process.
        return False

class TimedProcess(Process):
    """
    Helper to allow us to time a specific chunk of code and determine when
    it has reached a timeout.

    Also, this conveniently allows us to kill the whole thing if it locks up
    or takes too long, without requiring special coding in the target code.
    """

    daemon = True

    def __init__(self, max_seconds, time_type=c.MAX_TIME, fout=None, check_freq=1, *args, **kwargs):
        super(TimedProcess, self).__init__(*args, **kwargs)
        self.fout = fout or sys.stdout
        self.t0 = time.clock()
        self.t0_objective = time.time()
        self.max_seconds = float(max_seconds)
        self.t1 = None
        self.t1_objective = None
        # The number of seconds the process waits between checks.
        self.check_freq = check_freq
        self.time_type = time_type
        self._p = None
        self._process_times = {} # {pid:user_seconds}
        self._last_duration_seconds = None

    def terminate(self, signal=15, *args, **kwargs):
        """
        signal := 6=abrt, 9=kill, 15=term
        """
        if self.is_alive() and self._p:
            # Explicitly kill children since the default terminate() doesn't
            # seem to do this very reliably.
            for child in self._p.get_children():
                # Do one last time check.
                self._process_times[child.pid] = child.get_cpu_times().user
                os.system('kill -%i %i' % (signal, child.pid,))
            # Sum final time.
            self._process_times[self._p.pid] = self._p.get_cpu_times().user
            self._last_duration_seconds = sum(self._process_times.itervalues())
        os.system('kill -%i %i' % (signal, self._p.pid,))
        #return super(TimedProcess, self).terminate(*args, **kwargs)

    def get_duration_seconds_wall(self):
        if self.t1_objective is not None:
            return self.t1_objective - self.t0_objective
        return time.time() - self.t0_objective

    def get_duration_seconds_cpu(self):
        if self.t1 is not None:
            return self.t1 - self.t0
        return time.clock() - self.t0

    def get_duration_seconds_cpu_recursive(self):
        # Note, this calculation will consume much user
        # CPU time itself than simply checking clock().
        # Recommend using larger check_freq to minimize this.
        # Note, we must store historical child times because child
        # processes may die, causing them to no longer be included in
        # future calculations, possibly corrupting the total time.
        self._process_times[self._p.pid] = self._p.get_cpu_times().user
        children = self._p.get_children(recursive=True)
        for child in children:
            self._process_times[child.pid] = child.get_cpu_times().user
        #TODO:optimize by storing total sum and tracking incremental changes?
        return sum(self._process_times.itervalues())

    def get_cpu_usage_recursive(self, interval=1):
        usage = 0
        try:
            usage = self._p.get_cpu_percent(interval=interval)
            children = self._p.get_children(recursive=True)
            for child in children:
                try:
                    usage += child.get_cpu_percent(interval=interval)
                except psutil._error.NoSuchProcess:
                    pass
        except psutil._error.NoSuchProcess:
            pass
        return usage

    def get_duration_seconds_max(self):
        return max(
            self.get_duration_seconds_wall(),
            self.get_duration_seconds_cpu_recursive(),
        )

    def get_duration_seconds(self):
        """
        Retrieve the number of seconds this process has been executing for.

        If process was instantiated with objective=True, then the wall-clock
        value is returned.

        Otherwise the user-time is returned.
        If recursive=True is given, recursively finds all child-processed,
        if any, and includes their user-time in the total calculation.
        """
        if self.is_alive():
            if self.time_type == c.WALL_CLOCK_TIME:
                return self.get_duration_seconds_wall()
            elif self.time_type == c.CPU_TIME:
                return self.get_duration_seconds_cpu()
            elif self.time_type == c.RECURSIVE_CPU_TIME:
                return self.get_duration_seconds_cpu_recursive()
            elif self.time_type == c.MAX_TIME:
                return self.get_duration_seconds_max()
            else:
                raise NotImplementedError

    @property
    def is_expired(self):
        if not self.max_seconds:
            return False
        duration_seconds = self.get_duration_seconds()
        return duration_seconds >= self.max_seconds

    @property
    def seconds_until_timeout(self):
        return max(self.max_seconds - self.get_duration_seconds(), 0)

    def start(self, *args, **kwargs):
        super(TimedProcess, self).start(*args, **kwargs)
        self._p = psutil.Process(self.pid)

    def start_then_kill(self, verbose=True):
        """
        Starts and then kills the process if a timeout occurs.

        Returns true if a timeout occurred. False if otherwise.
        """
        self.start()
        timeout = False
        while 1:
            time.sleep(1)
            if verbose:
                self.fout.write('\r\t%.0f seconds until timeout.' \
                    % (self.seconds_until_timeout,))
                self.fout.flush()
            if not self.is_alive():
                break
            elif self.is_expired:
                if verbose:
                    six.print_('\nAttempting to terminate expired process %s...' % (self.pid,), file=self.fout)
                timeout = True
                self.terminate()
        self.t1 = time.clock()
        self.t1_objective = time.time()
        return timeout

def make_naive(dt, tz):
    if timezone.is_aware(dt):
        return timezone.make_naive(dt, tz)
    return dt

def make_aware(dt, tz):
    if settings.USE_TZ:
        if timezone.is_aware(dt):
            return dt
        else:
            return timezone.make_aware(dt, tz)
    else:
        if timezone.is_aware(dt):
            return timezone.make_naive(dt)
        else:
            return dt

def localtime(dt):
    dt = make_aware(dt, settings.TIME_ZONE)
    return dt
    
def write_lock(lock_file):
    lock_file.seek(0)
    lock_file.write(str(time.time()).encode('utf-8'))
    lock_file.flush()
    


from __future__ import print_function

from datetime import datetime
from distutils.version import StrictVersion

import django
DJANGO_VERSION = StrictVersion(django.get_version())

from django import forms
from django.conf import settings
from django.conf.urls import patterns, url
from django.contrib import admin
from django.core.management import get_commands
from django.core.urlresolvers import reverse, NoReverseMatch
from django.db import models
from django.forms import TextInput
if DJANGO_VERSION >= StrictVersion('1.8'):
    from django.forms.utils import flatatt
else:
    from django.forms.util import flatatt
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.utils.encoding import force_text
from django.http import HttpResponseRedirect, Http404, HttpResponse
from django.template.defaultfilters import linebreaks
from django.utils import dateformat, timezone
from django.utils.datastructures import MultiValueDict
from django.utils.formats import get_format
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.utils.text import capfirst
from django.utils.translation import (
    ungettext,
    ugettext_lazy as _
)

from chroniker.models import Job, Log, JobDependency, Monitor
from chroniker import utils
from chroniker.widgets import ImproveRawIdFieldsFormTabularInline

try:
    from admin_steroids.queryset import ApproxCountQuerySet
except ImportError:
    ApproxCountQuerySet = None

class HTMLWidget(forms.Widget):
    def __init__(self,rel=None, attrs=None):
        self.rel = rel
        super(HTMLWidget, self).__init__(attrs)
    
    def render(self, name, value, attrs=None):
        if self.rel is not None:
            key = self.rel.get_related_field().name
            obj = self.rel.to._default_manager.get(**{key: value})
            related_url = '../../../%s/%s/%d/' % (
                self.rel.to._meta.app_label,
                self.rel.to._meta.object_name.lower(),
                value)
            value = "<a href='%s'>%s</a>" % (related_url, escape(obj))
            
        final_attrs = self.build_attrs(attrs, name=name)
        return mark_safe("<div%s>%s</div>" % (
            flatatt(final_attrs),
            linebreaks(value)))

class JobDependencyInline(ImproveRawIdFieldsFormTabularInline):
    model = JobDependency
    extra = 1
    fk_name = 'dependent'
    
    readonly_fields = (
        'criteria_met',
    )
    
    raw_id_fields = (
        'dependee',
    )

class JobAdmin(admin.ModelAdmin):
    
    formfield_overrides = {
        models.CharField: {
            'widget': TextInput(attrs={'size':'100',})
        },
    }
    
    actions = (
        'enable_jobs',
        'disable_jobs',
        'run_selected_jobs',
        'toggle_enabled',
        'clear_stalled',
    )
    list_display = (
        'name',
        'last_run_with_link',
        'get_timeuntil',
        'get_frequency',
        'job_type',
        'hostname',
        'enabled',
        'check_is_complete',
        'is_fresh',
        #'is_monitor',
        'last_run_successful',
        'progress_percent_str',
        'estimated_completion_datetime_str',
        'run_button',
        'stop_button',
        'view_logs_button',
    )
    
    def job_type(self, obj=''):
        if not obj:
            return ''
        if obj.is_monitor:
            return 'monitor'
        return 'job'
    job_type.short_description = 'type'
    
    readonly_fields = (
        #'check_is_running',
        'check_is_complete',
        'view_logs_button',
        'last_run_successful',
        'last_heartbeat',
        'is_fresh',
        'last_run_start_timestamp',
        'last_run',
        'total_parts',
        'total_parts_complete',
        'progress_percent_str',
        'estimated_completion_datetime_str',
        'monitor_records',
        'current_hostname',
        'current_pid',
        'job_type',
        'is_due',
    )
    list_display_links = ('name', )
    list_filter = (
        'frequency',
        'enabled',
        'is_running',
        'last_run_successful',
        'hostname',
        'is_monitor',
    )
    filter_horizontal = ('subscribers',)
    fieldsets = (
        ('Job Details', {
            'classes': ('wide',),
            'fields': (
                'name',
                'command',
                'args',
                'raw_command',
                'hostname',
                'current_hostname',
                'current_pid',
            )
        }),
        ('Status', {
            'classes': ('wide',),
            'fields': (
                'is_running',
                'is_due',
                'is_fresh', 
                'last_run_successful',
                'total_parts',
                'total_parts_complete',
                'progress_percent_str',
                'estimated_completion_datetime_str',
                'last_heartbeat',
                'last_run_start_timestamp',
                'last_run',
            )
        }),
        ('Flags', {
            'classes': ('wide',),
            'fields': (
                'enabled',
                'force_run',
                'force_stop',
            )
        }),
        ('Logging', {
            'classes': ('wide',),
            'fields': (
                'view_logs_button',
                'log_stdout',
                'log_stderr',
                'maximum_log_entries',
            )
        }),
        ('Monitor', {
            'classes': ('wide',),
            'fields': (
                'is_monitor',
                'monitor_url',
                'monitor_error_template',
                'monitor_description',
                'monitor_records',
            )
        }),
        ('E-mail subscriptions', {
            'classes': ('wide',),
            'fields': (
                'subscribers',
                'email_errors_to_subscribers',
                'email_success_to_subscribers',
            )
        }),
        ('Frequency options', {
            'classes': ('wide',),
            'fields': (
                'frequency',
                'next_run',
                'params',
                'timeout_seconds',
            )
        }),
    )
    search_fields = (
        'name',
        'hostname',
        'current_hostname',
        'current_pid',
    )
    
    inlines = (
       JobDependencyInline,
    )
    
    class Media:
        js = ("chroniker/js/dygraph-combined.js",)
    
    def queryset(self, *args, **kwargs):
        qs = super(JobAdmin, self).queryset(*args, **kwargs)
        if ApproxCountQuerySet is not None:
            qs = qs._clone(klass=ApproxCountQuerySet)
        return qs
    
    def get_readonly_fields(self, request, obj=None):
        fields = list(self.readonly_fields)
        # Allow manual clearing of is_running if the cron job has become stuck.
#        if obj and obj.is_fresh():
#            fields.append('is_running')
        return fields
    
    def last_run_with_link(self, obj=None):
        if not obj or not obj.id:
            return ''
        format = get_format('DATETIME_FORMAT')
        value = None
        try:
            if obj.last_run is not None:
                value = utils.localtime(obj.last_run)
                value = capfirst(dateformat.format(value, format))
            log_id = obj.log_set.latest('run_start_datetime').id
            try:
                # Old way
                url = reverse('chroniker_log_change', args=(log_id,))
            except NoReverseMatch:
                # New way
                url = reverse('admin:chroniker_log_change', args=(log_id,))
            return '<a href="%s">%s</a>' % (url, value)
        except:
            return value
    last_run_with_link.admin_order_field = 'last_run'
    last_run_with_link.allow_tags = True
    last_run_with_link.short_description = 'Last run'
    
    def check_is_complete(self, obj=None):
        if not obj or not obj.id:
            return ''
        return not obj.check_is_running()
    check_is_complete.short_description = _('is complete')
    check_is_complete.boolean = True
    check_is_complete.admin_order_field = 'is_running'
    
    def get_timeuntil(self, obj=None):
        if not obj or not obj.id or not obj.next_run:
            return ''
        format = get_format('DATETIME_FORMAT')
        dt = obj.next_run
        dt = utils.localtime(dt)
        value = capfirst(dateformat.format(dt, format))
        return "%s<br /><span class='mini'>(%s)</span>" \
            % (value, obj.get_timeuntil())
    get_timeuntil.admin_order_field = 'next_run'
    get_timeuntil.allow_tags = True
    get_timeuntil.short_description = _('next scheduled run')
    
    def get_frequency(self, obj=None):
        if not obj or not obj.id:
            return ''
        freq = capfirst(obj.frequency.lower())
        if obj.params:
            return "%s (%s)" % (freq, obj.params)
        return freq
    get_frequency.admin_order_field = 'frequency'
    get_frequency.short_description = 'Frequency'
    
    def run_button(self, obj=None):
        if not obj or not obj.id:
            return ''
        url = '%d/run/?inline=1' % obj.id
        return '<a href="%s" class="button">Run</a>' % url
    run_button.allow_tags = True
    run_button.short_description = 'Run'
    
    def stop_button(self, obj=None):
        if not obj or not obj.id:
            return ''
        url = '%d/stop/?inline=1' % obj.id
        vars = dict(url=url, disabled='')
        if not obj.is_running:
            vars['disabled'] = 'disabled'
        s = ('<a href="%(url)s" class="button" %(disabled)s>Stop</a>') % vars
        return s
    stop_button.allow_tags = True
    stop_button.short_description = 'Stop'
    
    def view_logs_button(self, obj=None):
        if not obj or not obj.id:
            return ''
        q = obj.logs.all()
        url = utils.get_admin_changelist_url(Log)
        return ('<a href="%s?job=%d" target="_blank" class="button">View&nbsp;%i</a>') % \
            (url, obj.id, q.count())
    view_logs_button.allow_tags = True
    view_logs_button.short_description = 'Logs'
    
    def run_job_view(self, request, job_id):
        """
        Runs the specified job.
        """
        Job.objects.filter(id=job_id).update(
            force_run = True,
            force_stop = False,
        )
        self.message_user(
            request,
            _('Job %(job)s has been signalled to start running immediately.') \
                % {'job': job_id})
        if 'inline' in request.GET:
            redirect = request.path + '../../'
        else:
            redirect = request.REQUEST.get('next', request.path + "../")
        return HttpResponseRedirect(redirect)
    
    def stop_job_view(self, request, job_id):
        """
        Stop the specified job.
        """
        Job.objects.filter(id=job_id).update(
            force_run = False,
            force_stop = True,
        )
        self.message_user(
            request,
            _('Job %(job)s has been signalled to stop running immediately.') \
                % {'job': job_id})
        if 'inline' in request.GET:
            redirect = request.path + '../../'
        else:
            redirect = request.REQUEST.get('next', request.path + "../")
        return HttpResponseRedirect(redirect)
    
    def view_duration_graph(self, request, object_id):
        
        model = self.model
        opts = model._meta
        object_id = int(object_id)
        obj = self.get_object(request, object_id)
        
        q = obj.logs.all()
        q = q.order_by('run_start_datetime')
        q = q.only('duration_seconds', 'run_start_datetime')
        
        max_duration = q.aggregate(
            models.Max('duration_seconds')
        )['duration_seconds__max']
        
        errors = q.filter(success=False)
        
        media = self.media
        
        context = {
            'title': _('Change %s') % force_text(opts.verbose_name),
            #'adminform': adminForm,
            'object_id': object_id,
            'original': obj,
            'is_popup': False,
            'media': media,
            #'inline_admin_formsets': inline_admin_formsets,
            #'errors': helpers.AdminErrorList(form, formsets),
            'app_label': opts.app_label,
            'opts': opts,
            #'preserved_filters': self.get_preserved_filters(request),
            'q': q,
            'errors': errors,
            'max_duration': max_duration,
        }
        
        return render_to_response('admin/chroniker/job/duration_graph.html',
            context,
            context_instance=RequestContext(request))
    
    def get_urls(self):
        urls = super(JobAdmin, self).get_urls()
        my_urls = patterns('',
            url(r'^(.+)/run/$',
                self.admin_site.admin_view(self.run_job_view),
                name="chroniker_job_run"),
            url(r'^(.+)/stop/$',
                self.admin_site.admin_view(self.stop_job_view),
                name="chroniker_job_stop"),
            url(r'^(.+)/graph/duration/$',
                self.admin_site.admin_view(self.view_duration_graph),
                name='chroniker_job_duration_graph'),
        )
        return my_urls + urls
    
    def run_selected_jobs(self, request, queryset):
        #rows_updated = queryset.update(next_run=timezone.now())
        rows_updated = queryset.update(force_run=True)
        if rows_updated == 1:
            message_bit = "1 job was"
        else:
            message_bit = "%s jobs were" % rows_updated
        self.message_user(request, "%s successfully set to run." % message_bit)
    run_selected_jobs.short_description = "Force run selected jobs"
    
    def clear_stalled(self, request, queryset):
        reset_count = 0
        for job in queryset:
            if not job.is_fresh():
                reset_count += 1
                job.is_running = False
                job.save()
        self.message_user(request, 'Cleared %i stalled jobs.' % (reset_count,))
    clear_stalled.short_description = 'Mark the selected stalled %(verbose_name_plural)s as not running'
        
    def toggle_enabled(self, request, queryset):
        for row in queryset:
            row.enabled = not row.enabled
            row.save()
        rows_updated = queryset.count()
        if rows_updated == 1:
            message_bit = "1 job was toggled"
        else:
            message_bit = "%s jobs were toggled" % rows_updated
        self.message_user(request, message_bit)
    toggle_enabled.short_description = "Toggle enabled flag on selected jobs"
    
    def disable_jobs(self, request, queryset):
        queryset.update(enabled=False)
        rows_updated = queryset.count()
        if rows_updated == 1:
            message_bit = "1 job was toggled"
        else:
            message_bit = "%s jobs were toggled" % rows_updated
        self.message_user(request, message_bit)
    disable_jobs.short_description = "Disable selected jobs"
    
    def enable_jobs(self, request, queryset):
        queryset.update(enabled=True)
        rows_updated = queryset.count()
        if rows_updated == 1:
            message_bit = "1 job was toggled"
        else:
            message_bit = "%s jobs were toggled" % rows_updated
        self.message_user(request, message_bit)
    enable_jobs.short_description = "Enable selected jobs"
    
    def formfield_for_dbfield(self, db_field, **kwargs):
        request = kwargs.pop("request", None)
        
        # Add a select field of available commands
        if db_field.name == 'command':
            choices_dict = MultiValueDict()
            for command, app in get_commands().items():
                choices_dict.appendlist(app, command)
            
            choices = []
            for key in choices_dict.keys():
                #if str(key).startswith('<'):
                #    key = str(key)
                commands = choices_dict.getlist(key)
                commands.sort()
                choices.append([key, [[c,c] for c in commands]])
            
            choices.insert(0, ('', '--- None ---'))
            kwargs['widget'] = forms.widgets.Select(choices=choices)
            return db_field.formfield(**kwargs)
        
        kwargs['request'] = request
        return super(JobAdmin, self).formfield_for_dbfield(db_field, **kwargs)

class LogAdmin(admin.ModelAdmin):
    
    list_display = (
        'job_name',
        'run_start_datetime',
        'run_end_datetime',
        'duration_seconds',
        'duration_str',
        'job_success',
        'on_time',
        'hostname',
        #'stdout_sample',
        #'stderr_sample',
    )
    
    list_filter = (
        'success',
        'on_time',
    )
    
    search_fields = (
        'stdout',
        'stderr',
        'job__name',
        'job__command',
    )
    
    readonly_fields = (
        'run_start_datetime',
        'run_end_datetime',
        'duration_seconds',
        'stdout_long_sample',
        'stderr_long_sample',
        'stdout_link',
        'stderr_link',
        'duration_str',
        'hostname',
    )
    date_hierarchy = 'run_start_datetime'
    fieldsets = (
        (None, {
            'fields': (
                'job',
                'run_start_datetime',
                'run_end_datetime',
                'duration_seconds',
                'duration_str',
                'hostname',
            )
        }),
        ('Output', {
            'fields': (
                'stderr_link',
                'stderr_long_sample',
                'stdout_link',
                'stdout_long_sample',
            )
        }),
    )
    
    def queryset(self, *args, **kwargs):
        qs = super(LogAdmin, self).queryset(*args, **kwargs)
        qs = qs.only(
            'id',
            'run_start_datetime',
            'run_end_datetime',
            'duration_seconds',
            'success',
            'on_time',
        )
        if ApproxCountQuerySet is not None:
            qs = qs._clone(klass=ApproxCountQuerySet)
        return qs
    
    def stdout_link(self, obj):
        url = reverse("admin:chroniker_log_change", args=(obj.id,)) + 'stdout/'
        return '<a href="%s">Download</a>' % (url,)
    stdout_link.allow_tags = True
    stdout_link.short_description = 'Stdout full'
    
    def stderr_link(self, obj):
        url = reverse("admin:chroniker_log_change", args=(obj.id,)) + 'stderr/'
        return '<a href="%s">Download</a>' % (url,)
    stderr_link.allow_tags = True
    stderr_link.short_description = 'Stderr full'
    
    def view_full_stdout(self, request, log_id):
        log = Log.objects.get(id=log_id)
        resp = HttpResponse(
            log.stdout,
            content_type='application/x-download',
        )
        resp['Content-Disposition'] = 'filename=log-%s-stdout.txt' % (log_id,)
        return resp
    
    def view_full_stderr(self, request, log_id):
        log = Log.objects.get(id=log_id)
        resp = HttpResponse(
            log.stderr,
            content_type='application/x-download',
        )
        resp['Content-Disposition'] = 'filename=log-%s-stderr.txt' % (log_id,)
        return resp
    
    def get_urls(self):
        urls = super(LogAdmin, self).get_urls()
        my_urls = patterns('',
            (r'^(?P<log_id>[0-9]+)/stdout/?$',
                self.admin_site.admin_view(self.view_full_stdout)),
            (r'^(?P<log_id>[0-9]+)/stderr/?$',
                self.admin_site.admin_view(self.view_full_stderr)),
        )
        return my_urls + urls
    
    def job_name(self, obj):
      return obj.job.name
    job_name.short_description = _('Name')

    def job_success(self, obj):
        return obj.success
    job_success.short_description = _('OK')
    job_success.boolean = True
    job_success.admin_order_field = 'success'
    
    def has_add_permission(self, request):
        return False
    
    def formfield_for_dbfield(self, db_field, **kwargs):
        request = kwargs.pop("request", None)
        
        if isinstance(db_field, models.TextField):
            kwargs['widget'] = HTMLWidget()
            return db_field.formfield(**kwargs)
        
        if isinstance(db_field, models.ForeignKey):
            kwargs['widget'] = HTMLWidget(db_field.rel)
            return db_field.formfield(**kwargs)
        
        return super(LogAdmin, self).formfield_for_dbfield(db_field, **kwargs)

try:
    admin.site.register(Job, JobAdmin)
except admin.sites.AlreadyRegistered:
    pass

admin.site.register(Log, LogAdmin)

class MonitorAdmin(admin.ModelAdmin):
    list_display = (
        'name_str',
        'status',
        'monitor_records',
        'get_timeuntil',
        'enabled',
        'action_buttons',
    )
    list_filter = (
        'last_run_successful',
    )
    readonly_fields = (
        'name_str',
        'monitor_records',
        'action_buttons',
        'status',
        'get_timeuntil',
    )
    
    def get_timeuntil(self, obj):
        format = get_format('DATETIME_FORMAT')
        value = capfirst(dateformat.format(utils.localtime(obj.next_run), format))
        return "%s<br /><span class='mini'>(%s)</span>" \
            % (value, obj.get_timeuntil())
    get_timeuntil.admin_order_field = 'next_run'
    get_timeuntil.allow_tags = True
    get_timeuntil.short_description = _('next check')
    
    def get_actions(self, request):
        actions = super(MonitorAdmin, self).get_actions(request)
        del actions['delete_selected']
        return actions
    
    def has_delete_permission(self, request, obj=None):
        return False
    
    def has_add_permission(self, request):
        return False 
    
    def queryset(self, request):
        qs = super(MonitorAdmin, self).queryset(request)
        qs = qs.filter(is_monitor=True)
        qs = qs.order_by('name')
        return qs
    
    def name_str(self, obj):
        if obj.monitor_url:
            url = obj.monitor_url_rendered
            return '<a href="%s" target="_blank">%s</a>' % (url, obj.name)
        else:
            return obj.name
    name_str.short_description = 'Name'
    name_str.allow_tags = True
    
    def action_buttons(self, obj):
        buttons = []
        url = '%d/run/?inline=1' % obj.id
        buttons.append('<a href="%s" class="button">Check now</a>' % url)
        buttons.append('<a href="/admin/chroniker/job/%i/" target="_blank" class="button">Edit</a>' % (obj.id,))
        return ' '.join(buttons)
    action_buttons.allow_tags = True
    action_buttons.short_description = 'Actions'
    
    def status(self, obj):
        if obj.is_running:
            help_text = 'The monitor is currently being checked.'
            temp = '<img src="'+settings.STATIC_URL+'admin/img/icon-unknown.gif" alt="%(help_text)s" title="%(help_text)s" />'
        elif obj.last_run_successful:
            help_text = 'All checks passed.'
            temp = '<img src="'+settings.STATIC_URL+'admin/img/icon_success.gif" alt="%(help_text)s" title="%(help_text)s" />'
        else:
            help_text = 'Requires attention.'
            temp = '<img src="'+settings.STATIC_URL+'admin/img/icon_error.gif" alt="%(help_text)s" title="%(help_text)s" />'
        return temp % dict(help_text=help_text)
            
    status.allow_tags = True

    def changelist_view(self, request, extra_context=None):
        return super(MonitorAdmin, self).changelist_view(
            request, extra_context=dict(title='View monitors'))

    def run_job_view(self, request, pk):
        """
        Runs the specified job.
        """
        try:
            job = Job.objects.get(pk=pk)
        except Job.DoesNotExist:
            raise Http404
        # Rather than actually running the Job right now, we
        # simply force the Job to be run by the next cron job
        job.force_run = True
        job.save()
        self.message_user(
            request,
            _('The monitor "%(job)s" will be checked.') \
                % {'job': job})
        if 'inline' in request.GET:
            redirect = request.path + '../../'
        else:
            redirect = request.REQUEST.get('next', request.path + "../")
        return HttpResponseRedirect(redirect)
    
    def get_urls(self):
        urls = super(MonitorAdmin, self).get_urls()
        my_urls = patterns('',
            url(r'^(.+)/run/$',
                self.admin_site.admin_view(self.run_job_view),
                name="chroniker_job_run"),
#            url(r'^(.+)/stop/$',
#                self.admin_site.admin_view(self.stop_job_view),
#                name="chroniker_job_stop"),
        )
        return my_urls + urls

admin.site.register(Monitor, MonitorAdmin)



from __future__ import print_function

import time

from django.core.management.base import BaseCommand

from optparse import make_option

from chroniker.models import Job, Log

class Command(BaseCommand):
    help = 'Incrementally updates status, to help testing transaction ' + \
        'behavior on different database backends.'
    
    option_list = BaseCommand.option_list + (
        make_option('--seconds',
            dest='seconds',
            default=60,
            help='The number of total seconds to count up to.'),
        )
    
    def handle(self, *args, **options):
        seconds = int(options['seconds'])
        for i in xrange(seconds):
            Job.update_progress(total_parts=seconds, total_parts_complete=i)
            print('%i of %i' % (i, seconds))
            time.sleep(1)